=== modified file 'bzrlib/benchmarks/__init__.py'
--- bzrlib/benchmarks/__init__.py	2006-05-27 06:18:34 +0000
+++ bzrlib/benchmarks/__init__.py	2006-06-08 16:12:13 +0000
@@ -47,6 +47,7 @@
                    'bzrlib.benchmarks.bench_add',
                    'bzrlib.benchmarks.bench_checkout',
                    'bzrlib.benchmarks.bench_commit',
+                   'bzrlib.benchmarks.bench_osutils',
                    'bzrlib.benchmarks.bench_status',
                    'bzrlib.benchmarks.bench_transform',
                    'bzrlib.benchmarks.bench_workingtree',

=== modified file 'bzrlib/benchmarks/bench_add.py'
--- bzrlib/benchmarks/bench_add.py	2006-05-23 00:07:45 +0000
+++ bzrlib/benchmarks/bench_add.py	2006-06-08 16:12:13 +0000
@@ -24,6 +24,7 @@
     def test_one_add_kernel_like_tree(self):
         """Adding a kernel sized tree should be bearable (<5secs) fast.""" 
         self.make_kernel_like_tree()
-        # on roberts machine this originally took:  25936ms/32244ms
-        # after making smart_add use the parent_ie:  5033ms/ 9368ms
+        # on roberts machine: this originally took:  25936ms/32244ms
+        # after making smart_add use the parent_ie:   5033ms/ 9368ms
+        # plain os.walk takes 213ms on this tree
         self.time(self.run_bzr, 'add')

=== added file 'bzrlib/benchmarks/bench_osutils.py'
--- bzrlib/benchmarks/bench_osutils.py	1970-01-01 00:00:00 +0000
+++ bzrlib/benchmarks/bench_osutils.py	2006-06-08 16:12:13 +0000
@@ -0,0 +1,34 @@
+# Copyright (C) 2006 by Canonical Ltd
+#
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License version 2 as published by
+# the Free Software Foundation.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
+
+"""Tests for bzr osutils functions performance."""
+
+
+from bzrlib.benchmarks import Benchmark
+import bzrlib.osutils as osutils
+
+
+class WalkDirsBenchmark(Benchmark):
+
+    def test_walkdirs_kernel_like_tree(self):
+        """Walking a kernel sized tree is fast!(150ms)."""
+        self.make_kernel_like_tree()
+        # on roberts machine: this originally took:  157ms/4177ms
+        # plain os.walk takes 213ms on this tree
+        def dowalk():
+            for dirblock in osutils.walkdirs('.'):
+                if dirblock[0][1] == '.bzr':
+                    del dirblock[0]
+        self.time(dowalk)

=== modified file 'bzrlib/fetch.py'
--- bzrlib/fetch.py	2006-05-05 01:29:34 +0000
+++ bzrlib/fetch.py	2006-06-08 18:33:27 +0000
@@ -169,7 +169,16 @@
                     self.from_repository.get_transaction())
                 # we fetch all the texts, because texts do
                 # not reference anything, and its cheap enough
-                to_weave.join(from_weave, version_ids=required_versions)
+                to_weave.join(from_weave, version_ids=required_versions) 
+                # we dont need *all* of this data anymore, but we dont know
+                # what we do. This cache clearing will result in a new read 
+                # of the knit data when we do the checkout, but probably we
+                # want to emit the needed data on the fly rather than at the
+                # end anyhow.
+                # the from weave should know not to cache data being joined,
+                # but its ok to ask it to clear.
+                from_weave.clear_cache()
+                to_weave.clear_cache()
         finally:
             texts_pb.finished()
 

=== modified file 'bzrlib/osutils.py'
--- bzrlib/osutils.py	2006-06-08 15:08:04 +0000
+++ bzrlib/osutils.py	2006-06-08 16:12:13 +0000
@@ -16,16 +16,17 @@
 # along with this program; if not, write to the Free Software
 # Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 
-from shutil import copyfile
-from stat import (S_ISREG, S_ISDIR, S_ISLNK, ST_MODE, ST_SIZE,
-                  S_ISCHR, S_ISBLK, S_ISFIFO, S_ISSOCK)
 from cStringIO import StringIO
 import errno
 import os
+from os import listdir
 import re
 import sha
 import shutil
+from shutil import copyfile
 import stat
+from stat import (S_ISREG, S_ISDIR, S_ISLNK, ST_MODE, ST_SIZE,
+                  S_ISCHR, S_ISBLK, S_ISFIFO, S_ISSOCK)
 import string
 import sys
 import time
@@ -82,8 +83,10 @@
         return f
 
 
+_directory_kind = 'directory'
+
 _formats = {
-    stat.S_IFDIR:'directory',
+    stat.S_IFDIR:_directory_kind,
     stat.S_IFCHR:'chardev',
     stat.S_IFBLK:'block',
     stat.S_IFREG:'file',
@@ -91,17 +94,27 @@
     stat.S_IFLNK:'symlink',
     stat.S_IFSOCK:'socket',
 }
-def file_kind(f, _formats=_formats, _unknown='unknown', _lstat=os.lstat):
+
+
+def file_kind_from_stat_mode(stat_mode, _formats=_formats, _unknown='unknown'):
+    """Generate a file kind from a stat mode. This is used in walkdirs.
+
+    Its performance is critical: Do not mutate without careful benchmarking.
+    """
     try:
-        return _formats[_lstat(f).st_mode & 0170000]
+        return _formats[stat_mode & 0170000]
     except KeyError:
         return _unknown
 
 
+def file_kind(f, _lstat=os.lstat, _mapper=file_kind_from_stat_mode):
+    return _mapper(_lstat(f).st_mode)
+
+
 def kind_marker(kind):
     if kind == 'file':
         return ''
-    elif kind == 'directory':
+    elif kind == _directory_kind:
         return '/'
     elif kind == 'symlink':
         return '@'
@@ -788,3 +801,40 @@
         return
     if _validWin32PathRE.match(path) is None:
         raise IllegalPath(path)
+
+
+def walkdirs(top):
+    """Yield data about all the directories in a tree.
+    
+    This yields all the data about the contents of a directory at a time.
+    After each directory has been yielded, if the caller has mutated the list
+    to exclude some directories, they are then not descended into.
+    
+    The data yielded is of the form:
+    [(relpath, basename, kind, lstat, path_from_top), ...]
+
+    :return: an iterator over the dirs.
+    """
+    lstat = os.lstat
+    pending = []
+    _directory = _directory_kind
+    _listdir = listdir
+    pending = [("", "", _directory, None, top)]
+    while pending:
+        dirblock = []
+        currentdir = pending.pop()
+        # 0 - relpath, 1- basename, 2- kind, 3- stat, 4-toppath
+        top = currentdir[4]
+        if currentdir[0]:
+            relroot = currentdir[0] + '/'
+        else:
+            relroot = ""
+        for name in sorted(_listdir(top)):
+            abspath = top + '/' + name
+            statvalue = lstat(abspath)
+            dirblock.append ((relroot + name, name, file_kind_from_stat_mode(statvalue.st_mode), statvalue, abspath))
+        yield dirblock
+        # push the user specified dirs from dirblock
+        for dir in reversed(dirblock):
+            if dir[2] == _directory:
+                pending.append(dir)

=== modified file 'bzrlib/tests/test_osutils.py'
--- bzrlib/tests/test_osutils.py	2006-06-07 12:22:54 +0000
+++ bzrlib/tests/test_osutils.py	2006-06-08 16:12:13 +0000
@@ -220,3 +220,42 @@
     def test_split_with_carriage_returns(self):
         self.assertEqual(['foo\rbar\n'],
                          osutils.split_lines('foo\rbar\n'))
+
+
+class TestWalkDirs(TestCaseInTempDir):
+
+    def test_walkdirs(self):
+        tree = [
+            '.bzr',
+            '0file',
+            '1dir/',
+            '1dir/0file',
+            '1dir/1dir/',
+            '2file'
+            ]
+        self.build_tree(tree)
+        expected_dirblocks = [
+                [
+                    ('0file', '0file', 'file'),
+                    ('1dir', '1dir', 'directory'),
+                    ('2file', '2file', 'file'),
+                ],
+                [
+                    ('1dir/0file', '0file', 'file'),
+                    ('1dir/1dir', '1dir', 'directory'),
+                ],
+                [
+                ],
+            ]
+        result = []
+        found_bzrdir = False
+        for dirblock in osutils.walkdirs('.'):
+            if len(dirblock) and dirblock[0][1] == '.bzr':
+                # this tests the filtering of selected paths
+                found_bzrdir = True
+                del dirblock[0]
+            result.append(dirblock)
+
+        self.assertTrue(found_bzrdir)
+        self.assertEqual(expected_dirblocks,
+            [[line[0:3] for line in block] for block in result])

